const router = { params: {}, quary: {}, title: "", page: "" };
const store = {};
let navigate = () => {
  setTimeout(() => {
    navigate();
  });
};
function isInDocument(element) {
  var currentElement = element;
  while (currentElement && currentElement.parentNode) {
    if (currentElement.parentNode === document) {
      return true;
    } else if (currentElement.parentNode instanceof DocumentFragment) {
      currentElement = currentElement.parentNode.host;
    } else {
      currentElement = currentElement.parentNode;
    }
  }
  return false;
}
// Component
function Component(
  Name = String(),
  options = {
    Render({
      self = document.querySelector("*"),
      props = {},
      root = document.querySelector("*"),
    }) {},
    Script({
      self = document.querySelector("*"),
      props = {},
      root = document.querySelector("*"),
    }) {},
    Style({
      self = document.querySelector("*"),
      props = {},
      root = document.querySelector("*"),
    }) {},
    useOnce({
      self = document.querySelector("*"),
      props = {},
      root = document.querySelector("*"),
    }) {},
    shadowRoot: Boolean(true),
    useStore: Boolean(false),
  }
) {
  const { Render, Script, Style, useOnce, props } = options;
  window.customElements.define(
    Name.includes("-") ? Name : `${PREFIX}-${Name}`,
    class JamComponent extends HTMLElement {
      constructor() {
        super();
        let { shadowRoot } = options;
        if (shadowRoot == null || shadowRoot == undefined) {
          shadowRoot = true;
        }
        this.shadowRootMode = shadowRoot;
        if (this.shadowRootMode == true) {
          this.attachShadow({ mode: "open" });
        }
        if (options.useStore) {
          if (typeof options.useStore == "object") {
            options.useStore.forEach((storeName) => {
              store[`${storeName}.__Store_dep__`].push(this);
            });
          } else {
            store[`${options.useStore}.__Store_dep__`].push(this);
          }
        }
      }
      connectedCallback() {
        for (const key in props) {
          if (Object.hasOwnProperty.call(data, key)) {
            this.setAttribute(key, data[key]);
          }
        }
        const self = this;
        this.obs = new MutationObserver(function (mutations) {
          mutations.forEach(function (mutation) {
            if (mutation.type === "attributes") {
              if (mutation.target.shadowRootMode == true) {
                RenderComponent(
                  mutation.target.shadowRoot,
                  Render,
                  Script,
                  Style,
                  useOnce,
                  false
                );
              }
              if (mutation.target.shadowRootMode == false) {
                RenderComponent(
                  mutation.target,
                  Render,
                  Script,
                  Style,
                  useOnce,
                  false
                );
              }
            }
          });
        });

        this.obs.observe(this, {
          attributes: true,
        });
        if (this.shadowRootMode == true) {
          RenderComponent(
            this.shadowRoot,
            Render,
            Script,
            Style,
            useOnce,
            true
          );
        }
        if (this.shadowRootMode == false) {
          RenderComponent(this, Render, Script, Style, useOnce, true);
        }
        this.__update__ = () => {
          console.log("__update__");
          if (this.shadowRootMode == true) {
            RenderComponent(
              this.shadowRoot,
              Render,
              Script,
              Style,
              useOnce,
              false
            );
          }
          if (this.shadowRootMode == false) {
            RenderComponent(this, Render, Script, Style, useOnce, false);
          }
        };
      }
    }
  );
  return (data) => {
    const customElement = document.createElement(
      Name.includes("-") ? Name : `${PREFIX}-${Name}`
    );
    if (typeof data == "object") {
      for (const key in data) {
        if (Object.hasOwnProperty.call(data, key)) {
          customElement.setAttribute(key, data[key]);
        }
      }
    }
    return customElement;
  };
}
function DomParser(string = String()) {
  return new DOMParser().parseFromString(string, "text/html");
}
function RenderComponent(self, Render, Script, Style, useOnce, FirstTime) {
  //Element.attributes returns NamedNodeMap, to convert this into an Object
  const props = Object.assign(
    {},
    ...Array.from(
      self?.host ? self.host.attributes : self.attributes,
      ({ name, value }) => ({
        [name]: value,
      })
    )
  );
  self.innerHTML = ``;
  if (
    !Render({
      root: self?.host ? self.host : self,
      self: self.firstChild,
      props,
    })
  )
    return null;
  const elementRenderer = DomParser(
    Render({
      root: self?.host ? self.host : self,
      self: self.firstChild,
      props,
    })
  ).body.firstChild;
  self.append(elementRenderer);
  if (
    elementRenderer &&
    isInDocument(document.querySelector("slot[name=router]"))
  ) {
    RouterLink(elementRenderer);
  }
  if (
    Render({
      root: self?.host ? self.host : self,
      self: self.firstChild,
      props,
    }) &&
    typeof Style == "function"
  ) {
    self.append(
      DomParser(`<style>
        ${Style({
          root: self?.host ? self.host : self,
          self: self.firstChild,
          props,
        })}</style>
      `).head.querySelector("style")
    );
  }

  if (
    Render({
      root: self?.host ? self.host : self,
      self: self.firstChild,
      props: self?.host ? self.host.dataset : self.dataset,
    }) &&
    typeof Script == "function"
  ) {
    Script({
      root: self?.host ? self.host : self,
      self: self.firstChild,
      props,
    });
  }

  if (
    Render({
      root: self?.host ? self.host : self,
      self: self.firstChild,
      props,
    }) &&
    typeof useOnce == "function" &&
    FirstTime == true
  ) {
    useOnce({
      root: self?.host ? self.host : self,
      self: self.firstChild,
      props,
    });
  }
}
// Store
function CreateStore(StoreName = "app", value = {}) {
  store[StoreName] = new Proxy(value, {
    set(target, prop, value) {
      target[prop] = value;
      store[`${StoreName}.__Store_dep__`].forEach((component) => {
        if (isInDocument(component)) {
          component.__update__();
        }
      });
      return true;
    },
  });
  store[`${StoreName}.__Store_dep__`] = [];
  return store[StoreName];
}
function useStore(StoreName = "app") {
  return store[StoreName];
}

// Router
function CreateBrowserRouter({ exact = "/", _404, layout, Routes = [] }) {
  window.onpopstate = () => handleLocation({ exact, _404, layout, Routes });
  handleLocation({ exact, _404, layout, Routes });
  // Component("router-link", {
  //   Render({ data }) {
  //     return /*html*/ `
  //       <a href="${data.href}" >${data.text}</a>
  //     `;
  //   },
  //   Script({ root, self }) {
  //     self.addEventListener("click", (e) => {
  //       e.preventDefault();
  //       history.pushState(null, null, e.target.href);
  //       handleLocation({ exact, _404, layout, Routes });
  //     });
  //   },
  // });
  navigate = (url = "/") => {
    history.pushState(null, null, url);
    handleLocation({ exact, _404, layout, Routes });
  };
  return (url = "/") => {
    history.pushState(null, null, url);
    handleLocation({ exact, _404, layout, Routes });
  };
}
function useRouter() {
  return {
    params: router.params,
    quary: router.quary,
    page: router.page,
    title: router.title,
  };
}
function handleLocation({ exact, _404, layout, Routes }) {
  if (location.pathname != exact && location.pathname == "/") {
    history.pushState(null, null, exact);
    handleLocation({ exact, _404, layout, Routes });
  }
  const StaticRoute = Routes.find((route) => {
    const Route_url = ConvertURLToArray(route.path);
    const Location_url = ConvertURLToArray(location.pathname);
    return (
      Location_url.length == Route_url.length &&
      !route.path.includes(":") &&
      Location_url.join("") == Route_url.join("")
    );
  });

  if (StaticRoute) return RouterRender(StaticRoute, layout, {});
  let DynamicRouteFound = false;
  // const DynamicRoute = Routes.find((route) => {
  //   const Route_url = ConvertURLToArray(route.path);
  //   const Location_url = ConvertURLToArray(location.pathname);
  //   return Location_url.length == Route_url.length && route.path.includes(":");
  // });
  const DynamicRoutes = Routes.filter((route) => {
    return route.path.includes(":");
  });
  DynamicRoutes.forEach((route) => {
    const Route_url = ConvertURLToArray(route.path);
    const Dy_Route_url = Route_url.filter((param) => {
      return param.includes(":");
    });
    const Location_url = ConvertURLToArray(location.pathname);

    if (Route_url.length == Location_url.length) {
      let done = false;
      Route_url.forEach((param, index) => {
        if (!done && index + 1 == Route_url.length) {
          done = true;
          DynamicRouteFound = true;
          return RouterDynamicRender(
            route,
            ConvertURLToArray(route.path),
            ConvertURLToArray(location.pathname),
            layout,
            () => {
              history.replaceState(null, null, _404);
              handleLocation({ exact, _404, layout, Routes });
            }
          );
        }
        if (!param.includes(":")) {
          if (!done && param == Location_url[index]) {
            return true;
          } else {
            done = true;
          }
        }
      });
    }
  });
  // if (DynamicRoute)

  //   );
  if (!DynamicRouteFound && typeof _404 == "function") return _404();
  else if (!DynamicRouteFound && typeof _404 == "string") {
    history.replaceState(null, null, _404);
    handleLocation({ exact, _404, layout, Routes });
  }
}

function RouterRender(route, layout, url_params) {
  if (route.page) {
    setPage(route.page);
  }
  if (route.title) {
    setTitle(route.title);
  }
  setParams(url_params);
  setQuary();
  const RouterContainer = document.querySelector("slot[name=router]");
  RouterContainer.innerHTML = "";
  if (route.layout) {
    const LAYOUT = route.layout.cloneNode(true);
    RouterContainer.append(LAYOUT);
    if (LAYOUT.shadowRootMode) {
      LAYOUT.shadowRoot.querySelector("slot").replaceWith(route.render);
    } else {
      LAYOUT.querySelector("slot").replaceWith(route.render);
    }
  }
  if (!route.layout && layout) {
    const LAYOUT = layout.cloneNode(true);
    RouterContainer.append(LAYOUT);
    if (LAYOUT.shadowRootMode) {
      LAYOUT.shadowRoot.querySelector("slot").replaceWith(route.render);
    } else {
      LAYOUT.querySelector("slot").replaceWith(route.render);
    }
  }
  if (!route.layout && !layout) {
    RouterContainer.append(route.render);
  }
}
function RouterLink(root = document.querySelector("")) {
  root.querySelectorAll("a[link]").forEach((link) => {
    link.addEventListener("click", (e) => {
      e.preventDefault();
      navigate(link.href);
    });
  });
}
function RouterDynamicRender(route, Route_url, Location_url, layout, _404) {
  let url_params = {};
  const eq_url = CheckForEqBetweenTwoURLs(Route_url, Location_url);
  if (eq_url) {
    let value = true;
    Route_url.forEach((url, index) => {
      if (url.includes(":")) {
        Route_url[index] = Location_url[index];
        url_params[`${url.replace(":", "")}`] = Location_url[index];
      }
    });
    if (route) {
      RouterRender(route, layout, url_params);
    }
  } else {
    _404();
  }
}

function CheckForEqBetweenTwoURLs(Route_url, Location_url) {
  let value = true;
  Route_url.forEach((url, index) => {
    if (value) {
      if (!url.includes(":")) {
        if (url == Location_url[index]) {
          value = true;
        } else {
          value = false;
        }
      }
    }
  });
  return value;
}
function ConvertURLToArray(pathname) {
  let url = pathname.split("/");
  if (url[url.length - 1] == "") {
    url.pop();
  }
  url.shift();
  return url;
}
function setParams(params) {
  router.params = params;
}
function setQuary(QuaryString) {
  if (QuaryString) {
    const urlSearchParams = new URLSearchParams(QuaryString);
    const quary = Object.fromEntries(urlSearchParams.entries());
    router.quary = quary;
  } else {
    const urlSearchParams = new URLSearchParams(window.location.search);
    const quary = Object.fromEntries(urlSearchParams.entries());
    router.quary = quary;
  }
}
function setTitle(title = "") {
  router.title = title;
  document.title = title;
}
function setPage(page = "") {
  router.page = page;
}
