#!/usr/bin/env node
const prompts = require("prompts");
const { copy, readFile, writeFile } = require("fs-extra");
const path = require("path");
const template = [
  path.join(__dirname, "/template/default"),
  path.join(__dirname, "/template/tailwind"),
];
async function Main() {
  const response = await prompts([
    {
      type: "text",
      name: "name",
      message: `Enter The App Name`,
    },

    {
      type: "select",
      name: "value",
      message: "Pick a Template To Start",
      choices: [
        {
          title: "Bare Minimum",
          description: "Blank Soft application",
          value: 0,
        },
        {
          title: "Bare Minimum + Tailwindcss",
          description: "Blank Soft application + Tailwindcss",
          value: 1,
        },
      ],
      initial: 0,
    },
  ]);
  if (response.name.length == 0) {
    response.name = "soft-app";
  }
  copy(template[response.value], `${process.cwd()}/${response.name}`, {
    recursive: true,
    overwrite: false,
  })
    .then(() => {
      readFile(`${process.cwd()}/${response.name}/package.json`, {
        encoding: "utf-8",
      })
        .then((data) => {
          const packageJson = JSON.parse(data);
          packageJson.name = response.name;
          const packageJsonData = JSON.stringify(packageJson, null, 2);
          writeFile(
            `${process.cwd()}/${response.name}/package.json`,
            packageJsonData
          )
            .then(() => {
              readFile(path.join(__dirname, "/dist/Soft.esm.min.js"))
                .then((data) => {
                  writeFile(
                    `${process.cwd()}/${
                      response.name
                    }/src/static/js/Soft.esm.js`,
                    data
                  )
                    .then(() => {
                      console.clear();
                      console.log("Your App Is Ready !!");
                      console.log(``);
                      console.log(
                        `Thank You For Choosing Soft For Your Application, And Have Fun :)`
                      );
                      console.log(``);
                      console.log(`  - cd ${response.name}`);
                      console.log("");
                      console.log(`  - npm install`);
                      console.log("");
                      console.log(`  - npm run dev`);
                      console.log("");
                      if (response.value == 1) {
                        console.log(`   npm run watch:css`);
                      }
                    })
                    .catch((err) => {
                      console.log(err);
                    });
                })
                .catch((err) => {
                  console.log(err);
                });
            })
            .catch((err) => {
              console.log(err);
            });
        })
        .catch((err) => {
          console.log(err);
        });
    })
    .catch((err) => {
      console.log(err);
    });
}
Main();
